from .calculations import normalize as VecNorm
from .calculations import point_distance as VecDistance
from .calculations import vector_angle as VecAngle
from .calculations import vector_inverse as VecInv
from .calculations import position_between_xz as PosBetweenXZ
from .calculations import sort_nearest_points_from_on_plane_y as SortPointsFromPlaneY
from .calculations import calculate_reflection_angle as RefAngle
from .converters import outdoor_traced_result_to_line as OutdoorResultToLines
from .converters import dbm_to_mw as dBmTomW
from .converters import mw_to_dbm as mWTodBm


import unittest
import radio_gyms


class TestVersion(unittest.TestCase):

    def test_version(self):
        """Check Version from radio_gyms"""
        self.assertEqual(radio_gyms.__version__, "0.2.0")




from numpy.typing import NDArray
from . import Transmitter


class Cell:
    transmitter: Transmitter = None
    position: NDArray = None

    def __init__(self, position, transmit_power):
        self.transmitter = Transmitter(position, transmit_power)



from typing import List, Tuple

import numpy as np

from ..utils import VecDistance, RefAngle, VecNorm, VecAngle, SortPointsFromPlaneY, dBmTomW, mWTodBm
from ..utils.constants import LIGHT_SPEED


class TheoreticalOutdoorModel:
    """
    Theoretical Outdoor model consist of free-space loss, reflection loss, and knife-edge diffraction.
    This theoretical outdoor model can be used for ultra-high frequency up to 100 GHz.

    The ray tracing result can be calculated for loss and signal propagation delay
    """
    result = None

    def __init__(self, result, tx_power_dbm: float):
        self.result = result
        self.tx_power_dBm = tx_power_dbm

    @staticmethod
    def calculate_free_space_loss(tx_pos, rx_pos, frequency, wave_speed=LIGHT_SPEED):
        tx_pos = np.array(tx_pos)
        rx_pos = np.array(rx_pos)
        distance = VecDistance(tx_pos, rx_pos)
        wave_length = wave_speed / frequency
        space_loss = (4 * np.pi * distance / wave_length) ** 2
        return 10 * np.log10(space_loss)

    @staticmethod
    def calculate_single_reflection_loss(tx_pos, rx_pos, ref_pos, frequency,
                                         tx_medium_permittivity=1,
                                         ref_medium_permittivity=5.31,
                                         polar='TM',
                                         wave_speed=LIGHT_SPEED):
        tx_pos = np.array(tx_pos)
        rx_pos = np.array(rx_pos)
        ref_pos = np.array(ref_pos)
        # calculate coefficient
        assert polar == 'TM' or polar == 'TE'

        angle_1 = RefAngle(tx_pos, ref_pos, rx_pos)
        angle_2 = np.arcsin(np.sqrt(tx_medium_permittivity) * np.sin(angle_1) / np.sqrt(ref_medium_permittivity))
        bias = np.sqrt(np.abs(tx_medium_permittivity / ref_medium_permittivity) * np.sin(angle_1))
        if bias >= 1:
            coefficient = 1
        else:
            sqrt_n1 = np.sqrt(tx_medium_permittivity)
            sqrt_n2 = np.sqrt(ref_medium_permittivity)
            cos_1 = np.cos(angle_1)
            cos_2 = np.cos(angle_2)
            if polar == 'TM':
                coefficient = (sqrt_n2 * cos_1 - sqrt_n1 * cos_2) / (sqrt_n2 * cos_1 + sqrt_n1 * cos_2)
            elif polar == 'TE':
                coefficient = (sqrt_n1 * cos_1 - sqrt_n2 * cos_2) / (sqrt_n1 * cos_1 + sqrt_n2 * cos_2)
            else:
                raise TypeError("invalid polar")
        # calculate loss
        distance = VecDistance(tx_pos, ref_pos) + VecDistance(ref_pos, rx_pos)
        wave_length = wave_speed / frequency
        loss = (4 * np.pi * distance / (coefficient * wave_length)) ** 2
        return 10 * np.log10(loss)

    @staticmethod
    def calculate_knife_edge_diffraction(tx_pos, rx_pos, frequency: float, edges: List, wave_speed=LIGHT_SPEED):
        tx_pos = np.array(tx_pos)
        rx_pos = np.array(rx_pos)
        edges = [np.array(edge) for edge in edges]

        def calculate_v(left_pos, right_pos, edge_pos):
            tx_to_rx_dir = VecNorm(right_pos - left_pos)
            rx_to_tx_dir = -tx_to_rx_dir
            tx_to_edge_dir = VecNorm(edge_pos - left_pos)
            rx_to_edge_dir = VecNorm(edge_pos - right_pos)
            angle_tx = VecAngle(tx_to_edge_dir, tx_to_rx_dir)
            angle_rx = VecAngle(rx_to_edge_dir, rx_to_tx_dir)
            wave_length = wave_speed / frequency
            r_1 = VecDistance(left_pos, edge_pos)
            r_2 = VecDistance(edge_pos, right_pos)
            s_1 = np.cos(angle_tx) * r_1
            s_2 = np.cos(angle_rx) * r_2
            h = np.sin(angle_tx) * r_1
            return h * np.sqrt((2 * (s_1 + s_2)) / (wave_length * r_1 * r_2))

        def calculate_c(v: float):
            return 6.9 + 20 * np.log10(np.sqrt((v - 0.1) ** 2 + 1) + v - 0.1)

        def calculate_correction(near_tx_pos, center_pos, near_rx_pos, near_tx_c_, center_c_, near_rx_c_):
            tx_pos_on_y = np.copy(tx_pos)
            tx_pos_on_y[1] = 0
            near_tx_on_y = np.copy(near_tx_pos)
            near_tx_on_y[1] = 0
            center_on_y = np.copy(center_pos)
            center_on_y[1] = 0
            near_rx_on_y = np.copy(near_rx_pos)
            near_rx_on_y[1] = 0
            rx_pos_on_y = np.copy(rx_pos)
            rx_pos_on_y[1] = 0
            d1 = VecDistance(tx_pos_on_y, near_tx_on_y)
            d2 = VecDistance(near_tx_on_y, center_on_y)
            d3 = VecDistance(center_on_y, near_rx_on_y)
            d4 = VecDistance(near_rx_on_y, rx_pos_on_y)

            cos_1 = np.sqrt((d1 * (d3 + d4)) / ((d1 + d2) * (d2 + d3 + d4)))
            cos_2 = np.sqrt((d4 * (d1 + d2)) / ((d3 + d4) * (d2 + d3 + d1)))

            correct_1 = (6 - center_c_ + near_tx_c_) * cos_1
            correct_2 = (6 - center_c_ + near_rx_c_) * cos_2
            return correct_1 + correct_2

        def sort_edge_by_v(unsorted_edges):
            return sorted(unsorted_edges, key=lambda edge: calculate_v(tx_pos, rx_pos, edge))

        def single_edge_loss(single_edge):
            v = calculate_v(tx_pos, rx_pos, single_edge)
            diff_loss = calculate_c(v)
            return diff_loss

        def double_edge_loss(double_edges):
            near_tx_edge_i = double_edges[0]
            near_tx_v_i = calculate_v(tx_pos, rx_pos, near_tx_edge_i)
            near_rx_edge_i = double_edges[1]
            near_rx_v_i = calculate_v(tx_pos, rx_pos, near_rx_edge_i)
            if near_tx_v_i > near_rx_v_i:
                main_c = calculate_c(near_tx_v_i)
                support_c = calculate_c(calculate_v(near_tx_edge_i, rx_pos, near_rx_edge_i))
            else:
                main_c = calculate_c(near_rx_v_i)
                support_c = calculate_c(calculate_v(tx_pos, near_rx_edge_i, near_tx_edge_i))
            diff_loss = main_c + support_c
            return diff_loss

        def more_edge_loss(more_edges):
            if len(more_edges) == 3:
                near_tx_edge = sorted_edges[0]
                center_edge = sorted_edges[1]
                near_rx_edge = sorted_edges[2]
            else:
                top_edges = sort_edge_by_v(sorted_edges)[:3]
                sorted_top_edge = SortPointsFromPlaneY(tx_pos, top_edges)
                near_tx_edge = sorted_top_edge[0]
                center_edge = sorted_top_edge[1]
                near_rx_edge = sorted_edges[2]

            near_tx_v = calculate_v(tx_pos, rx_pos, near_tx_edge)
            center_v = calculate_v(tx_pos, rx_pos, center_edge)
            near_rx_v = calculate_v(tx_pos, rx_pos, near_rx_edge)
            max_v = min([near_tx_v, center_v, near_rx_v])
            if near_tx_v == max_v:
                main_c = calculate_c(near_tx_v)
                support_c_1 = calculate_c(calculate_v(near_tx_edge, near_rx_edge, center_edge))
                support_c_2 = calculate_c(calculate_v(center_edge, rx_pos, near_rx_edge))
            elif near_rx_v == max_v:
                main_c = calculate_c(near_rx_v)
                support_c_1 = calculate_c(calculate_v(tx_pos, center_edge, near_tx_edge))
                support_c_2 = calculate_c(calculate_v(near_tx_edge, near_rx_edge, center_edge))
            else:
                main_c = calculate_c(center_v)
                support_c_1 = calculate_c(calculate_v(tx_pos, center_edge, near_tx_edge))
                support_c_2 = calculate_c(calculate_v(center_edge, rx_pos, near_rx_edge))
            if np.isnan(support_c_1):
                support_c_1 = 0
            if np.isnan(support_c_2):
                support_c_2 = 0
            near_tx_c = calculate_c(near_tx_v)
            center_c = calculate_c(center_v)
            near_rx_c = calculate_c(near_rx_v)
            total_correct = calculate_correction(near_tx_edge, center_edge, near_rx_edge,
                                                 near_tx_c, center_c, near_rx_c)
            diff_loss = main_c + support_c_1 + support_c_2 - total_correct
            return diff_loss

        edge_n = len(edges)

        loss = TheoreticalOutdoorModel.calculate_free_space_loss(tx_pos, rx_pos, frequency, wave_speed)
        sorted_edges = SortPointsFromPlaneY(tx_pos, edges)
        if edge_n == 1:
            single_loss = single_edge_loss(edges[0])
            loss += single_loss
        elif edge_n == 2:
            loss += double_edge_loss(sorted_edges)
        elif edge_n >= 3:
            loss += more_edge_loss(sorted_edges)
        return loss

    def calculate_max_received_power(self, frequency: float = 2.4e9, wave_speed=LIGHT_SPEED) -> float:
        tx_pos = self.result['tx_pos']
        rx_pos = self.result['rx_pos']
        receive_power_dbms = []
        free_space_loss = TheoreticalOutdoorModel.calculate_free_space_loss(tx_pos, rx_pos,
                                                                            frequency, wave_speed)

        if self.result['direct']:
            receive_power_dbm = self.tx_power_dBm - free_space_loss
            receive_power_dbms.append(receive_power_dbm)
        else:
            knife_edges = self.result['roof_edges']
            diffraction_loss = TheoreticalOutdoorModel.calculate_knife_edge_diffraction(tx_pos,
                                                                                        rx_pos,
                                                                                        frequency,
                                                                                        knife_edges,
                                                                                        wave_speed)
            receive_power_dbm = self.tx_power_dBm - diffraction_loss
            receive_power_dbms.append(receive_power_dbm)

        for ref_pos in self.result['reflections']['single']:
            reflected_loss = TheoreticalOutdoorModel.calculate_single_reflection_loss(tx_pos,
                                                                                      rx_pos,
                                                                                      ref_pos,
                                                                                      frequency,
                                                                                      wave_speed=wave_speed)
            reflected_received_power_dbm = self.tx_power_dBm - reflected_loss
            receive_power_dbms.append(reflected_received_power_dbm)

        received_power_mw = 0
        for receive_power_dbm in receive_power_dbms:
            received_power_mw += dBmTomW(receive_power_dbm)
        return mWTodBm(received_power_mw)

    @staticmethod
    def calculate_signal_delay(tx_pos, rx_pos, points=[], wave_speed=LIGHT_SPEED):
        tx_pos = np.array(tx_pos)
        rx_pos = np.array(rx_pos)
        points = [np.array(point) for point in points]
        distance = 0
        if len(points) == 0:
            distance += VecDistance(tx_pos, rx_pos)
        else:
            left_pos = tx_pos
            for i in range(len(points)):
                right_pos = points[i]
                distance += VecDistance(left_pos, right_pos)
                left_pos = points[i]
            right_pos = rx_pos
            distance += VecDistance(left_pos, right_pos)
        delay = distance / wave_speed
        return delay

    def calculate_signal_impulses(self, freq: float = 2.4e9,
                                  wave_speed: float = LIGHT_SPEED) -> List[Tuple[float, float]]:
        impulses = []
        tx_pos = self.result['tx_pos']
        rx_pos = self.result['rx_pos']
        if self.result['direct']:
            rev_power_dbm = self.tx_power_dBm - self.calculate_free_space_loss(tx_pos, rx_pos, freq, wave_speed)
            rev_delay = self.calculate_signal_delay(tx_pos, rx_pos)
            impulse = {'strength': rev_power_dbm,
                       'delay': rev_delay}
            impulses.append(impulse)
        else:
            knife_edges = self.result['roof_edges']
            diffraction_loss = TheoreticalOutdoorModel.calculate_knife_edge_diffraction(tx_pos,
                                                                                        rx_pos,
                                                                                        freq,
                                                                                        knife_edges,
                                                                                        wave_speed=wave_speed)
            rev_power_dbm = self.tx_power_dBm - diffraction_loss
            rev_delay = self.calculate_signal_delay(tx_pos, rx_pos, knife_edges)
            impulse = {'strength': rev_power_dbm,
                       'delay': rev_delay}
            impulses.append(impulse)
        for ref_pos in self.result['reflections']['single']:
            reflected_loss = TheoreticalOutdoorModel.calculate_single_reflection_loss(tx_pos,
                                                                                      rx_pos,
                                                                                      ref_pos,
                                                                                      freq,
                                                                                      wave_speed=wave_speed)
            rev_power_dbm = self.tx_power_dBm - reflected_loss
            rev_delay = self.calculate_signal_delay(tx_pos, rx_pos, [ref_pos])
            impulse = {'strength': rev_power_dbm,
                       'delay': rev_delay}
            impulses.append(impulse)
        return impulses



import os
import unittest
import numpy as np

from radio_gyms.engines import Tracer
from radio_gyms.engines.ray_tracer.triangle import Triangle
from radio_gyms.engines.ray_tracer.box import Box
from numpy.typing import NDArray

POZNAN_OBJ_PATH = os.path.join(os.getcwd(), "assets", "models", "poznan.obj")


class TestTracer(unittest.TestCase):

    def test_triangle(self):
        a: NDArray = np.array([5, 0, -3])
        b: NDArray = np.array([-5, 0, -2])
        c: NDArray = np.array([0, -4, 4])
        test_triangle = Triangle(a, b, c)

        ray_pos = np.array([0, 15, 0])
        ray_dir = np.array([0, -1, 0])
        ray = (ray_pos, ray_dir)
        result = test_triangle.is_intersect(ray)
        self.assertGreater(result, 0)
        ray_pos2 = np.array([0, 15, 0])
        ray_dir2 = np.array([0, 1, 0])
        ray2 = (ray_pos2, ray_dir2)
        result = test_triangle.is_intersect(ray2)
        self.assertLess(result, 0)

    def test_box(self):
        a: NDArray = np.array([5, 0, -3])
        b: NDArray = np.array([-5, 0, -2])
        c: NDArray = np.array([0, -4, 4])
        test_triangle = Triangle(a, b, c)
        box = Box([test_triangle])

        ray_pos = np.array([0, 0, 0])
        ray_dir = np.array([0, -1, 0])
        ray = (ray_pos, ray_dir)
        result = box.is_intersect(ray)
        self.assertEqual(result, True)

        ray_pos2 = np.array([0, -1000, 0])
        ray_dir2 = np.array([1, 0, 0])
        ray2 = (ray_pos2, ray_dir2)
        result2 = box.is_intersect(ray2)
        self.assertEqual(result2, False)

    def test_tracer_direct(self):
        tracer = Tracer(POZNAN_OBJ_PATH)
        tx_pos = np.array([0, 15, 0])
        rx_pos = np.array([-30, 1.5, 45])
        result1 = tracer.trace_outdoor(tx_pos, rx_pos)
        self.assertEqual(result1["direct"], False)
        tx_pos2 = np.array([0, -100, 0])
        rx_pos2 = np.array([0, -1, 0])
        result2 = tracer.trace_outdoor(tx_pos2, rx_pos2)
        self.assertEqual(result2["direct"], True)

    def test_tracer_reflect(self):
        tracer = Tracer(POZNAN_OBJ_PATH)
        tx_pos = np.array([0, 15, 0])
        rx_pos = np.array([-30, 1.5, 45])
        result = tracer.trace_single_reflect(tx_pos, rx_pos)
        self.assertEqual(len(result), 2)
        ref_1 = result[0]
        ref_2 = result[1]
        self.assertAlmostEqual(ref_1[0], -28.94, 1)
        self.assertAlmostEqual(ref_1[1], 4.22, 1)
        self.assertAlmostEqual(ref_1[2], 62.39, 1)
        self.assertAlmostEqual(ref_2[0], -70.80, 1)
        self.assertAlmostEqual(ref_2[1], 7.04, 1)
        self.assertAlmostEqual(ref_2[2], 15.22, 1)

    def test_roof_diffraction(self):
        tracer = Tracer(POZNAN_OBJ_PATH)
        tx_pos = np.array([0, 15, 0])
        rx_pos = np.array([-30, 1.5, 45])
        result = tracer.trace_roof_edges(tx_pos, rx_pos)
        self.assertEqual(len(result), 1)
        self.assertAlmostEqual(result[0][0], -19.24, 1)
        self.assertAlmostEqual(result[0][1], 8.56, 1)
        self.assertAlmostEqual(result[0][2], 28.86, 1)



import numpy as np

from numpy.typing import NDArray

from ..utils import VecNorm, VecDistance
from . import Receiver


class Pedestrian:
    receiver: Receiver = None
    position: NDArray = None
    speed = None
    destination = None
    walking = False

    def __init__(self, position, speed=1.4):
        self.position = np.array(position)
        self.speed = speed
        self.receiver = Receiver(position)

    def walk(self, delta_time):
        if not self.walking:
            return

        direction = VecNorm(self.position - self.destination)
        step_distance = self.speed * delta_time
        self.position += direction * self.speed * step_distance

        if VecDistance(self.position, self.destination) <= step_distance:
            self.walking = False



import os
from unittest import TestCase

from radio_gyms.engines import Tracer
from radio_gyms.simulations import OldtownWalk

POZNAN_OBJ_PATH = os.path.join(os.getcwd(), "assets", "models", "poznan.obj")

class TestOldtownWalkSimulation(TestCase):
    def test_impulse(self):
        tracer = Tracer(POZNAN_OBJ_PATH)
        simulation = OldtownWalk(tracer, 1, 30)

        simulation.update(1)
        result = simulation.get_impulse_list()
        print(result)



import sys

EPSILON: float = 0.0000001

MIN_FLT: float = sys.float_info.min
MAX_FLT: float = sys.float_info.max

MIN_ROOF_EDGE_DISTANCE = 0.3
ROOF_MIN_ANGLE = 0.0017
ROOF_MAX_SCAN = 20

LIGHT_SPEED = 299792458



from typing import Sequence, Tuple, List
import numpy as np
from numpy.typing import NDArray
from .triangle import Triangle
from ...utils.constants import MAX_FLT, MIN_FLT
from ...utils import VecInv


class Box:
    min_bound: NDArray = None
    max_bound: NDArray = None
    triangles: List[Triangle] = None
    child_left: 'Box' = None
    child_right: 'Box' = None
    alone: bool = False

    def __init__(self, triangles: List[Triangle]):
        self.triangles = [triangle for triangle in triangles]
        if len(self.triangles) == 1:
            self.alone = True

        min_x, min_y, min_z = MAX_FLT, MAX_FLT, MAX_FLT
        max_x, max_y, max_z = MIN_FLT, MIN_FLT, MIN_FLT
        for triangle in self.triangles:
            min_x = min(min_x, triangle.min_x)
            min_y = min(min_y, triangle.min_y)
            min_z = min(min_z, triangle.min_z)
            max_x = max(max_x, triangle.max_x)
            max_y = max(max_y, triangle.max_y)
            max_z = max(max_z, triangle.max_z)
        self.min_bound = np.array([min_x, min_y, min_z])
        self.max_bound = np.array([max_x, max_y, max_z])

    def is_intersect(self, ray: Tuple[NDArray, NDArray]) -> bool:
        """
        check if the input ray intersects with the box

        :param ray:
        :return: true/false
        """
        # Optimized from the algorithm: "fast branch-less ray-bounding box intersection"
        ray_pos: NDArray = np.array(ray[0])
        ray_inv_dir: NDArray = VecInv(np.array(ray[1]))

        min_bound_check = (self.min_bound - ray_pos) * ray_inv_dir
        max_bound_check = (self.max_bound - ray_pos) * ray_inv_dir
        t1, t3, t5 = min_bound_check
        t2, t4, t6 = max_bound_check
        t_min = max(max(min(t1, t2), min(t3, t4)), min(t5, t6))
        t_max = min(min(max(t1, t2), max(t3, t4)), max(t5, t6))
        return t_max >= t_min



from typing import Tuple
import numpy as np

class Transmitter:
    position: Tuple[float, float, float]
    transmit_power_dbm: float = 0

    def __init__(self, position: Tuple[float, float, float], transmit_power_dbm: float):
        self.position = position
        self.transmit_power_dbm = transmit_power_dbm

    def set_tx_power(self, transmit_power_dbm):
        self.transmit_power_dbm = transmit_power_dbm



import numpy as np

from typing import List

from ..engines import Tracer
from ..objects import Pedestrian, Cell
from ..models import TheoreticalOutdoorModel as OutdoorModel

class OldtownWalk:
    tracer: Tracer = None
    cells: List[Cell] = []
    pedestrians: List[Pedestrian] = []
    step: int = 0

    def __init__(self, tracer, cell_n, pedestrian_n):
        self.step = 0
        self.tracer = tracer
        self.pedestrian_position_limit = {
            'min_x': tracer.min_bound[0],
            'max_x': tracer.max_bound[0],
            'min_y': 0.5,
            'max_y': 1.8,
            'min_z': tracer.min_bound[2],
            'max_z': tracer.max_bound[2],
        }
        self.cell_position_limit = {
            'min_x': tracer.min_bound[0],
            'max_x': tracer.max_bound[0],
            'min_y': 2,
            'max_y': tracer.max_bound[1],
            'min_z': tracer.min_bound[2],
            'max_z': tracer.max_bound[2],
            'min_tx': 15,
            'max_tx': 20
        }
        self.cells, self.pedestrians = [], []
        self.generate_cell_randomly(pedestrian_n)
        self.generate_pedestrian_randomly(cell_n)

    def generate_cell_randomly(self, cell_n):
        for i in range(cell_n):
            position = self.random_outdoor_position(self.cell_position_limit)
            tx_power = np.random.uniform(low=self.cell_position_limit['min_tx'],
                                         high=self.cell_position_limit['max_tx'], size=1)[0]
            cell = Cell(position, tx_power)
            self.cells.append(cell)

    def generate_pedestrian_randomly(self, pedestrian_n):
        for i in range(pedestrian_n):
            position = self.random_outdoor_position(self.pedestrian_position_limit)
            pedestrian = Pedestrian(position)
            self.pedestrians.append(pedestrian)

    def random_outdoor_position(self, limit):
        outdoor = False
        while not outdoor:
            random_x = np.random.uniform(low=limit['min_x'],
                                         high=limit['max_x'], size=1)[0]
            random_y = np.random.uniform(low=limit['min_y'],
                                         high=limit['max_y'], size=1)[0]
            random_z = np.random.uniform(low=limit['min_z'],
                                         high=limit['max_z'], size=1)[0]
            new_position = np.array([random_x, random_y, random_z])
            if self.tracer.is_outdoor(new_position):
                outdoor = True
        return new_position

    def repurpose_pedestrian(self):
        for pedestrian in self.pedestrians:
            if not pedestrian.walking:
                found = False
                try_count = 0
                while not found:
                    try_count += 1
                    new_destination = self.random_outdoor_position(self.pedestrian_position_limit)
                    if self.tracer.direct_path(pedestrian.position, new_destination):
                        pedestrian.destination = new_destination
                        found = True
        return

    def update(self, delta_time):
        for pedestrian in self.pedestrians:
            pedestrian.walk(delta_time)

    def get_impulse_list(self):
        cell_results = []
        for cell_id, cell in enumerate(self.cells):
            tx_pos = cell.transmitter.position
            cell_result = {
                'id': cell_id,
                'pedestrians': []
            }
            for pedestrian_id, pedestrian in enumerate(self.pedestrians):
                rx_pos = pedestrian.receiver.position
                traced_result = self.tracer.trace_outdoor(tx_pos, rx_pos)
                model = OutdoorModel(traced_result, cell.transmitter.transmit_power_dbm)
                impulses = model.calculate_signal_impulses()
                max_rev_power = model.calculate_max_received_power()
                pedestrian_result = {
                    'id': pedestrian_id,
                    'impulses': impulses,
                    'rev_power': max_rev_power
                }
                cell_result['pedestrians'].append(pedestrian_result)
            cell_results.append(cell_result)
        return cell_results


from unittest import TestCase
from radio_gyms.models import TheoreticalOutdoorModel as model


class TestModelOutdoor(TestCase):
    def test_free_space_loss(self):
        wave_speed = 3e8
        test_set = [
            {'tx':  [0, 5, 0], 'rx': [34, 1.5, -59], 'freq': 2.3e9, 'expect': 76.35},
            {'tx': [0, 5, 0], 'rx': [24, 1.5, 0], 'freq': 2.3e9, 'expect': 67.3719},
            {'tx': [0, 5, 0], 'rx': [42, 1.5, -9], 'freq': 2.3e9, 'expect': 72.3650},
            {'tx': [0, 5, 0], 'rx': [-5, 1.5, 16], 'freq': 2.3e9, 'expect': 64.3487}
        ]
        for test in test_set:
            calculated_loss_db = model.calculate_free_space_loss(test['tx'], test['rx'],
                                                                 test['freq'], wave_speed)
            self.assertAlmostEqual(test['expect'], calculated_loss_db, 2)

    def test_reflection_loss(self):
        wave_speed = 3e8
        freq = 2.3e9
        test_set = [
            {'tx': [0, 5, 0], 'rx': [34, 1.5, -59],
             'ref': [36.1217, 2.834910, -30.0909],
             'polar': 'TE',
             'freq': freq,
             'expect': 81.093},
            {'tx': [0, 5, 0], 'rx': [34, 1.5, -59],
             'ref': [26.1838, 0.010512, -45.4366],
             'polar': 'TE',
             'freq': freq,
             'expect': 77.170},
            {'tx': [0, 5, 0], 'rx': [24, 1.5, 0],
             'ref': [64.1040, 2.846930, 0.0000],
             'polar': 'TE',
             'freq': freq,
             'expect': 88.1103}
        ]
        for test in test_set:
            ref_loss_db = model.calculate_single_reflection_loss(test['tx'], test['rx'],
                                                                 test['ref'], test['freq'],
                                                                 tx_medium_permittivity=1.0003,
                                                                 ref_medium_permittivity=5.3,
                                                                 wave_speed=wave_speed, polar=test['polar'])
            self.assertAlmostEqual(test['expect'], ref_loss_db, 1)

    def test_single_diffraction_loss(self):
        wave_speed = 3e8
        freq = 2.3e9
        test_set = [
            {
                'tx': [0,5,	0],
                'rx': [62, 1.5, -22],
                'freq': freq,
                'edge': [49.2605, 32.1573, -17.4795],
                'expect': 115.540
            },
            {
                'tx': [0, 5, 0],
                'rx': [45, 1.5, -36	],
                'freq': freq,
                'edge': [35.3552, 26.6196, -28.2842],
                'expect': 113.503
            },
            {
                'tx': [0, 5, 0],
                'rx': [61, 1.5, -19],
                'freq': freq,
                'edge': [51.8878, 31.8698, -16.1618],
                'expect': 115.229
            },
        ]
        for test in test_set:
            tx_pos = test['tx']
            rx_pos = test['rx']
            wave_freq = test['freq']
            edges = [test['edge']]
            expected_loss = test['expect']
            total_loss_db = model.calculate_knife_edge_diffraction(tx_pos, rx_pos,
                                                                   wave_freq, edges,
                                                                   wave_speed=wave_speed)
            self.assertAlmostEqual(expected_loss, total_loss_db, 2)

    def test_double_diffraction_loss(self):
        wave_speed = 3e8
        freq = 2.3e9
        test_set = [
            {
                'tx': [0, 5, 0],
                'rx': [20, 1.5, 0],
                'freq': freq,
                'edge': [
                    [10, 7, 0],
                    [15, 10, 0],
                ],
                'expect': 120.261
            },
        ]
        for test in test_set:
            tx_pos = test['tx']
            rx_pos = test['rx']
            wave_freq = test['freq']
            edges = test['edge']
            expected_loss = test['expect']
            total_loss_db = model.calculate_knife_edge_diffraction(tx_pos, rx_pos,
                                                                   wave_freq, edges,
                                                                   wave_speed=wave_speed)
            single_edge_loss = model.calculate_knife_edge_diffraction(tx_pos, rx_pos,
                                                                      wave_freq, [edges[0]],
                                                                      wave_speed=wave_speed)
            self.assertGreater(total_loss_db, single_edge_loss)
            self.assertAlmostEqual(expected_loss, total_loss_db, 2)

    def test_triple_diffraction_loss(self):
        wave_speed = 3e8
        freq = 2.3e9
        test_set = [
            {
                'tx': [0, 5, 0],
                'rx': [30, 1.5, 0],
                'freq': freq,
                'edge': [
                            [10, 7, 0],
                            [15, 10, 0],
                            [20, 7, 0]
                         ],
                'expect': 130.51
            },
        ]
        for test in test_set:
            tx_pos = test['tx']
            rx_pos = test['rx']
            edges = test['edge']
            freq = test['freq']
            expected_loss = test['expect']
            total_loss_db = model.calculate_knife_edge_diffraction(tx_pos, rx_pos,
                                                                   freq, edges,
                                                                   wave_speed=wave_speed)
            self.assertGreater(total_loss_db, 120.261)
            self.assertAlmostEqual(total_loss_db, expected_loss, 1)

    def test_signal_delay(self):
        test_set = [
            {
                'tx': [0, 0, 0],
                'rx': [0, 3e8, 0],
                'points': [],
                'expect': 1
            },
            {
                'tx': [0, 0, 0],
                'rx': [0, 6e8,0],
                'points': [[0, 3e8, 0], [0, 0e8, 0], [0, 3e8, 0]],
                'expect': 4
            }
        ]
        for test in test_set:
            tx_pos = test['tx']
            rx_pos = test['rx']
            points = test['points']
            expected = test['expect']
            delay = model.calculate_signal_delay(tx_pos, rx_pos, points, wave_speed=3e8)
            self.assertAlmostEqual(expected, delay, 0)


"""
This file is currently not used, it was replaced by pywavefront to avoid the complexity.
"""
import ctypes
from typing import List
from pyglet.gl import glUseProgram, GLchar, glCreateShader,\
    GL_VERTEX_SHADER, glShaderSource, glCompileShader, GL_FRAGMENT_SHADER,\
    glCreateProgram, glAttachShader, glLinkProgram, glGenBuffers, GLuint,\
    GL_ARRAY_BUFFER, glBindBuffer, GLfloat, GL_STATIC_DRAW, glBufferData, \
    glVertexAttribPointer, GL_FLOAT, GL_FALSE, glEnableVertexAttribArray, \
    glUniform3fv, glGetUniformLocation, glUniform4fv

import numpy as np

from ..utils import VecNorm

class VisualTriangle:
    shader = None

    def __init__(self, point_a: List, point_b: List, point_c: List, color_a: List):
        normal = VecNorm(np.cross(np.array(point_b) - np.array(point_a), np.array(point_c) - np.array(point_a)))
        self.triangle = [*point_a, *normal,
                         *point_b, *normal,
                         *point_c, *normal,]
        self.color = color_a
        self.vertex_shader_binary = b"""
        #version 330
        in layout(location=0) vec3 position;
        in layout(location=1) vec3 normal;
        
        // out vec3 FragPos;
        out vec3 newNormal;
        
        //uniform mat4 model;
        //uniform mat4 view;
        //uniform mat4 projection;
        
        void main(){
            // FragPos = vec3(model*vec4(position, 1.0));
            newNormal = normal;
            
            //gl_Position = projection * view * model * Vec4(position, 1.0f);
            gl_Position = Vec4(position, 1.0f);
        }
        """

        self.vertex_shader_binary = b"""
        #version 330
        layout(location = 0) in vec3 position;
        layout(location = 1) in vec3 color;

        out vec3 newColor;

        void main()
        {
            gl_Position = vec4(position, 1.0f);
            newColor = color;
        }
        """

        self.fragment_shader_binary = b"""
        #version 330
        out vec4 FragColor;
        
        in vec3 newNormal;
        
        // uniform vec3 lightPos;
        // uniform vec3 lightColor;
        // uniform vec3 objectColor;
        
        void main(){
            /*float ambientStrength = 0.1;
            vec3 ambient = ambientStrength * lightColor;

            vec3 norm = normalize(Normal);
            vec3 lightDir = normalize(lightPos - FragPos);
            float diff = max(dot(norm, lightDir), 0.0);
            vec3 diffuse = diff * lightColor;

            vec3 result = (ambient + diffuse) * objectColor;
            FragColor = vec4(result, 1.0);*/
            outColor = vec4(1.0f, 0.0f, 0.0f 1.0f);
        }
        """
        self.fragment_shader_binary = b"""
        #version 330
        in vec3 newColor;

        out vec4 outColor;

        void main()
        {
            outColor = vec4(newColor, 1.0f);
        }
        """

        vertex_buffer = ctypes.create_string_buffer(self.vertex_shader_binary)
        c_vertex = ctypes.cast(ctypes.pointer(ctypes.pointer(vertex_buffer)),
                               ctypes.POINTER(ctypes.POINTER(GLchar)))
        vertex_shader = glCreateShader(GL_VERTEX_SHADER)
        glShaderSource(vertex_shader, 1, c_vertex, None)
        glCompileShader(vertex_shader)

        fragment_buffer = ctypes.create_string_buffer(self.fragment_shader_binary)
        c_fragment = ctypes.cast(ctypes.pointer(ctypes.pointer(fragment_buffer)),
                               ctypes.POINTER(ctypes.POINTER(GLchar)))
        fragment_shader = glCreateShader(GL_FRAGMENT_SHADER)
        glShaderSource(fragment_shader, 1, c_fragment, None)
        glCompileShader(fragment_shader)

        self.shader = glCreateProgram()
        glAttachShader(self.shader, vertex_shader)
        glAttachShader(self.shader, fragment_shader)
        glLinkProgram(self.shader)

        glUseProgram(self.shader)

        vbo = GLuint(0)
        glGenBuffers(1, vbo)

        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        glBufferData(GL_ARRAY_BUFFER, 72, (GLfloat * len(self.triangle))(*self.triangle), GL_STATIC_DRAW)

        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 24, ctypes.c_void_p(0))
        glEnableVertexAttribArray(0)

        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 24, ctypes.c_void_p(12))
        glEnableVertexAttribArray(1)

    def update(self, view, projection, lightPos, lightColor, objectColor):
        model = np.ones((4,4))
